import obonet as on


def make_id_to_term_dictionary(obo_file_name: str) -> dict[str, str]:
    """
    uses obonet to parse the given obo file and returns a dictionary
    containing ontology ID to onotology term mappings

    :param obo_file_name:   string containing the path to the obo file to read

    :return:                dictionary containing ID to term mapping with ID as key and term as values
    """

    obo_graph = on.read_obo(obo_file_name)
    return {id_: data.get('name') for id_, data in obo_graph.nodes(data=True)}


def map_id_to_term(ont_id: str, ontologies: list[dict]) -> str:
    """
    maps a given ontology ID to it's human readable term. Returns an empty string if ID is not
    contained in any of the given ID to term mappings

    :param ont_id:      string containing the ontology ID to be mapped to its term
    :param ontologies:  list of id_to_term dictionaries for all used ontologies

    :return:            ontology term or empty string if ID is not found in any of the ontologies
    """
    term = None
    for ontology in ontologies:
        if not ont_id in ontology:
            continue

        term = ontology[ont_id]
    
    return term



import gc

import numpy as np

from scipy.sparse import csr_matrix 
from typing import Union, Iterable

# code in this file is a copy of code in scipy.sparse._construct
# somewhat memoy optimized for concatenating csr_matrices
def get_index_dtype(
    arrays: Iterable[np.ndarray] = (), 
    maxval: Union[float, None] = None, 
    check_contents: bool = False
) -> Union[np.int32, np.int64]:
    """
    Based on input (integer) arrays `a`, determine a suitable index data
    type that can hold the data in the arrays.

    Parameters
    ----------
    arrays : tuple of array_like
        Input arrays whose types/contents to check
    maxval : float, optional
        Maximum value needed
    check_contents : bool, optional
        Whether to check the values in the arrays and not just their types.
        Default: False (check only the types)

    Returns
    -------
    dtype : dtype
        Suitable index data type (int32 or int64)

    """

    int32min = np.int32(np.iinfo(np.int32).min)
    int32max = np.int32(np.iinfo(np.int32).max)

    # not using intc directly due to misinteractions with pythran
    dtype = np.int32 if np.intc().itemsize == 4 else np.int64
    if maxval is not None:
        maxval = np.int64(maxval)
        if maxval > int32max:
            dtype = np.int64

    if isinstance(arrays, np.ndarray):
        arrays = (arrays,)

    for arr in arrays:
        arr = np.asarray(arr)
        if not np.can_cast(arr.dtype, np.int32):
            if check_contents:
                if arr.size == 0:
                    # a bigger type not needed
                    continue
                elif np.issubdtype(arr.dtype, np.integer):
                    maxval = arr.max()
                    minval = arr.min()
                    if minval >= int32min and maxval <= int32max:
                        # a bigger type not needed
                        continue

            dtype = np.int64
            break

    return dtype


def _compressed_sparse_csr_stack(blocks: list[csr_matrix]) -> csr_matrix:
    """
    Stacking fast path for vstack for CSR
    """
    axis = 0
    other_axis = 1
    data = np.concatenate([b.data for b in blocks])
    constant_dim = blocks[0].shape[other_axis]
    idx_dtype = get_index_dtype(
        arrays = [b.indptr for b in blocks],
        maxval = max(data.size, constant_dim)
    )
    indices = np.empty(data.size, dtype=idx_dtype)
    indptr = np.empty(sum(b.shape[axis] for b in blocks) + 1, dtype=idx_dtype)
    last_indptr = idx_dtype(0)
    sum_dim = 0
    sum_indices = 0
    for b in blocks:
        if b.shape[other_axis] != constant_dim:
            raise ValueError(f'incompatible dimensions for axis {other_axis}')
        indices[sum_indices:sum_indices+b.indices.size] = b.indices
        sum_indices += b.indices.size
        idxs = slice(sum_dim, sum_dim + b.shape[axis])
        indptr[idxs] = b.indptr[:-1]
        indptr[idxs] += last_indptr
        sum_dim += b.shape[axis]
        last_indptr += b.indptr[-1]
    
    del blocks
    gc.collect()

    indptr[-1] = last_indptr
    return csr_matrix((data, indices, indptr), shape=(sum_dim, constant_dim))
    

def bmat(matrices: list[csr_matrix], dtype=None):
    # stack along rows (axis 0):
    A = _compressed_sparse_csr_stack(matrices)
    if dtype is not None:
        A = A.astype(dtype)
    return A



import h5py
import gc
import scipy
import logging

import pandas as pd
import numpy as np
import anndata as ad

from os import PathLike
from . import concurrency
from . import sparse
from typing import Union, Iterable, Any


logging.basicConfig(
    format = '%(threadName)s: %(asctime)s-%(levelname)s-%(message)s',
    datefmt = '%Y-%m-%d %H:%M:%S',
    level = logging.INFO
)


def parse_table(archs4_file: Union[str, PathLike], root_key: str, table_key: str) -> dict[str, Any]:
    """
    parses the archs4 h5 file and extracts the data at /root_key/table_key into a dictionary
    see https://maayanlab.cloud/archs4/help.html for more infos on the available options of
    root_key and table_key
    
    :param archs4_file:   path to the archs4 h5 file to parse
    :param root_key:      root key to use for parsing the file metadata
    :param table_key:     table key to use for parsing the file metadata
    
    :return:              dictionary containing the metadata with column names as keys and column contents as values
    """
    archs4 = h5py.File(archs4_file, 'r')
    
    data = dict()
    table = archs4[root_key][table_key]
    for key in table.keys():
        column_data_piece = table[key][0]
        if isinstance(column_data_piece, np.number):
            column_data = table[key][:]
        
        else:
            column_data = [x.decode('utf-8') for x in table[key][:]]
        
        data[key] = column_data
    
    return data


def filter_table_data(data: dict, retain_keys: Iterable[str]) -> dict[str, Any]:
    """
    filters the data dictionary to retain only the keys specified by retain_keys
    
    :param data:          dictionary as returned by parse_table
    :param retain_keys:   list of data keys to retain
    
    :return:              filtered data dictionary
    """
    retain_data = {
        key: data[key].copy() for key in retain_keys
    }
    return retain_data


def get_srx_samn_from_items(relation_items: Iterable[str]) -> tuple[str, str]:
    """
    takes list of strings generated from an item in the 'relation' column
    of the /meta/samples table in the archs4 h5 file and extracts the SRA and
    BioSample accessions from it
    
    :param relation_items:   list of strings of the format 'key: value'
    
    :return:                 sra and bisample accession as strings
    """
    srx, samn = None, None
    for item in relation_items:
        if not item:
            continue
            
        key, value = item.split(': ')
            
        if key == 'SRA':
            srx = value.split('=')[-1]
        
        if key == 'BioSample':
            samn = value.split('/')[-1]
        
    return srx, samn


def extract_srx_and_samn_accessions(relation_data: Iterable[str]) -> dict[str, list[str]]:
    """
    takes the full 'relation' column of the archs4 h5 /meta/samples table and extracts
    SRA and BioSample accessions for each of the items contained
    
    :param relation_data:  list of strings of the format 'key1: value1, key2: value2, ...'
    
    :return:               dictionary containing keys srx_accession, biosample_accession with lists of these accessions as values
    """
    srx_samn_accessions = {
        key: [] for key in ['srx_accession', 'biosample_accession']
    }
    for relation in relation_data:
        relation_items = relation.split(',')
        srx, samn = get_srx_samn_from_items(
            relation_items
        )
        srx_samn_accessions['srx_accession'].append(srx)
        srx_samn_accessions['biosample_accession'].append(samn)
    
    return srx_samn_accessions


def get_filtered_sample_metadata(archs4_file: Union[str, PathLike], keys_to_retain: Iterable[str]) -> pd.DataFrame:
    """
    parses the archs4 h5 file and returns a pandas.DataFrame of with all columns of the /meta/samples table 
    specified in keys_to_retain plus extracted SRA and BioSamples accessions (columns srx_accession and biosample_accession)
    
    :param archs4_file:       path to the archs4 file to parse
    :param keys_to_retain:    list of keys in the /meta/samples table to retain
    
    :return:                  pandas.DataFrame containing the respective metadata + SRA and BioSample accessions
    """
    data = parse_table(
        archs4_file,
        'meta',
        'samples'
    )

    filtered_data = filter_table_data(
        data, 
        keys_to_retain
    )

    srx_samn_accessions = extract_srx_and_samn_accessions(
        filtered_data['relation']
    )

    srx_samn_table = pd.DataFrame().from_dict(
        srx_samn_accessions,
        orient = 'columns'
    )

    table = pd.DataFrame().from_dict(
        filtered_data, 
        orient = 'columns'
    )

    table = pd.concat(
        [table, srx_samn_table],
        axis = 1
    )

    del data, filtered_data, srx_samn_accessions
    gc.collect()
    
    return table


def get_gene_metadata_table(file: Union[PathLike, str]) -> pd.DataFrame:
    """
    read gene/transcript metadata from ARCHS4 file

    :param file:        string denoting the path to the ARCHS4 file

    :return:            pandas.DataFrame containing the gene metadata
    """
    with h5py.File(file, 'r') as f:
        if "genes" in list(f["meta"].keys()):
            table_key = 'genes'

        elif "transcripts" in list(f["meta"].keys()):
            table_key = 'meta/transcripts'
            
        else:
            raise Exception("error in gene/transcript meta data")
        
        gene_metadata = {}
        for key in f['meta'][table_key].keys():
            gene_metadata[key] = [x.decode('utf-8') for x in f['meta'][table_key][key]]

    gene_metadata = pd.DataFrame(gene_metadata)
    gene_metadata.index = ad.utils.make_index_unique(gene_metadata.symbol)
    
    return gene_metadata


# adapted from archs4py as this would not install due to issues with Python 3.12 and numpy requirement
def load_data(
    file: Union[PathLike, str], 
    sample_idx: Iterable[int], 
    gene_idx: Iterable[int] = [], 
    n_processes: int = 1
) -> tuple[scipy.sparse.csr_matrix, pd.DataFrame]:
    """
    Retrieve gene expression data from a specified file for the given sample and gene indices.

    :param file:                the file path or object containing the data.
    :param sample_idx:          a list of sample indices to retrieve expression data for.
    :param gene_idx:            a list of gene indices to retrieve expression data for. Defaults to an empty list (return all).
    :param n_processes:         integer denoting the number of concurrent processes to use for data loading

    :return:                    scipy.sparse.csr_matrix containing expression data and a pandas.DataFrame containing gene metadata
    """
    sample_idx = sorted(sample_idx)
    gene_idx = sorted(gene_idx)
    genes_h5_path = get_gene_name_column(file)
    with h5py.File(file, "r") as f:
        genes = np.array([x.decode("UTF-8") for x in np.array(f[genes_h5_path])])
        
    if len(gene_idx) == 0:
        gene_idx = list(range(len(genes)))

    logging.info(f'start loading {len(sample_idx)} samples using {n_processes} process(es)')
    data = concurrency.process_data_in_chunks(
        sample_idx,
        load_samples_in_consecutive_blocks,
        n_processes = n_processes,
        file = file,
        gene_idx = gene_idx
    )

    # scipy.sparse.vstack uses a lot of memory when concatenating the data chunks
    # estimated to double/triple the size of the originally loaded data
    # we use an optimized version of it that is only supposed to concatenate csr_matrices
    logging.info('concatenate matrices')
    return sparse.bmat(data), get_gene_metadata_table(file).iloc[gene_idx, :]


def consecutive(
    sequence_array: np.ndarray[int], 
    *additional_arrays_to_split: np.ndarray[Any]
) -> list[tuple[np.ndarray]]:
    """
    finds all sequences of consecutive elements in a numpy array
    and splits the input arrays accordingly

    :param sequence_array:                  numpy.ndarray containig a sequence of integers
    :param *additional_arrays_to_split:     any number of additional arrays to split into same chunks as sequence array

    :return:                                list of tuples of chunks of sequence_array and additional arrays            
    """
    split_idx = np.where(np.diff(sequence_array) != 1)[0]+1
    arrays = [sequence_array, *additional_arrays_to_split]
    return [np.split(array, split_idx) for array in arrays]


def get_gene_name_column(file: Union[PathLike, str]) -> str:
    """
    detect the conatined gene/transcript names and return corresponding h5 path

    :param file:        string denoting the path to the ARCHS4 file

    :return:            string giving the h5 path to gene/transcript names
    """
    with h5py.File(file) as f:
        if "genes" in list(f["meta"].keys()):
            if "gene_symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/gene_symbol"
            elif "symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/symbol"
        elif "transcripts" in list(f["meta"].keys()):
            if "ensembl_id" in list(f["meta/trancripts"].keys()):
                return "meta/trancripts/ensembl_id"
        else:
            raise Exception("error in gene/transcript meta data")


def read_sample_data(
    file: Union[PathLike, str], 
    sample_idx: Iterable[int], 
    gene_idx: Iterable[int]
) -> scipy.sparse.csr_matrix:
    """
    reads a given block of consecutive samples from the ARCHS4 file

    :param file:                string denoting the path to the ARCHS4 file
    :param sample_idx:          iterable of integers denoting the samples to load from file
    :param gene_idx:            iterable of integers containing the indexes of genes to retain

    :return:                    scipy.sparse.csr_matrix containing the loaded expression data
    """
    with h5py.File(file, "r") as f:
        dense_expression = f["data/expression"][:, sample_idx][gene_idx, :]
        sparse_expression = scipy.sparse.csr_matrix(dense_expression.T)
        sparse_expression.eliminate_zeros()

        del dense_expression
        gc.collect()

    return sparse_expression
    

# this is a wrapper to ensure compliance with API
def load_samples_in_consecutive_blocks(
    sample_idx: Iterable[int], 
    file: Union[PathLike, str], 
    gene_idx: Iterable[int], 
    **kwargs
) -> scipy.sparse.csr_matrix:
    """
    load the data in consecutive blocks of samples to minimize IO

    :param sample_idx_gsms:     iterable of tuples containing the index of the sample and its accession
    :param file:                string denoting the path to the ARCHS4 file
    :param gene_idx:            iterable of integers containing the indexes of genes to retain
    :param **kwargs:            Any other keyword arguments. This does not have an effect but is just to comply to
                                'process_data_in_chunks' API

    :return:                    scipy.sparse.csr_matrix containing expression data
    """ 
    data = [
        read_sample_data(file, consecutive_idx, gene_idx) 
        for consecutive_idx in consecutive(sample_idx)[0]
    ]
    return sparse.bmat(data)


def samples(
    file: Union[PathLike, str], 
    sample_metadata: pd.DataFrame,
    n_processes: int = 1
) -> ad.AnnData:
    """
    retrieve samples corresponding to sample_metadata.index from ARCHS4 and pack them
    into an AnnData object. sample_metadata.index has to be GEO GSM accessions. This function
    will only return data for accessions found in ARCHS4

    :param file:                string denoting path to ARCHS4 file
    :param sample_metadata:     pandas.DataFrame indexed by GEO GSM accessions containing additional data
                                a subset of this will be saved to the obs of the generated AnnData object
    :param n_processes:         number of processes to use for loading the data from ARCHS4

    :return:                    AnnData containing expression and metadata of all GSM accessions contained in ARCHS4
    """
    sample_ids = set(sample_metadata.index)
    with h5py.File(file, "r") as f:
        samples = [x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])]

        sample_idx = [i for i,x in enumerate(samples) if x in sample_ids]
        if len(sample_idx) == 0:
            raise ValueError('No samples selected. Make sure to use valid GSM accessions!')
        
        gsm_ids = np.array([x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])])[sample_idx]

    sparse_expression, gene_metadata = load_data(file, sample_idx, n_processes=n_processes)

    logging.info('generating AnnData object')
    return ad.AnnData(X = sparse_expression, var = gene_metadata, obs = sample_metadata.loc[gsm_ids, :])
    


import multiprocessing as mp
import itertools as it

from functools import partial
from os import PathLike
from typing import Iterable, Callable, Union, TextIO, Any


def write(file_handle: TextIO, uid_list: list[list[str]]) -> None:
    """
    writes the uids in uid_list to file_handle and flushes it afterwards
    
    :param file_handle:   IO stream generated by functions like open
    :param uid_list:      list of lists of accession uid pairs as strings
    
    :return:              None
    """
    for uid_map in uid_list:
        file_handle.write(
            '\t'.join(uid_map) + '\n'
        )

    file_handle.flush()


def write_to_file_multiprocess(filename: Union[PathLike, str], uid_list: list[list[str]], filelock: mp.Manager().Lock) -> None:
    """
    writes the given list of uids to a file that is accesses from a
    concurrent process.
    
    :param filename:   path to the file to write the list to
    :param uid_list:   list of lists of accession uid pairs as strings
    :param filelock:   Lock object that can be used with multiprocessing imap
    
    :return:           None
    """
    with filelock:
        with open(filename, 'a') as outfile:
            write(outfile, uid_list)
            
            
def write_to_file_singleprocess(filename: Union[PathLike, str], uid_list: list[list[str]], *args):
    """
    writes the given list of uids to a file. single threaded equivalent to write_to_file_multiprocess
    
    :param
    """
    with open(filename, 'a') as outfile:
        write(outfile, uid_list)
        

def multiprocess_map(chunks: Iterable, func: Callable, n_processes: int, **kwargs) -> None:
    """
    takes an iterable containing chunks of a whole and a function to process these chunks and uses n_processes to do it
    
    :param chunks:        Iterable containing chunks of a whole that need to be processed concurrently
    :param func:          function to process the chunks with
    :param n_processes:   number of concurrent processes to use for processing
    :param **kwargs:      any keyword arguments to pass to func
    
    :return:              None
    """
    with mp.Pool(n_processes) as p:
        # lock needs to be a managed one otherwise passing it to the pool will fail
        filelock = mp.Manager().Lock()
        
        map_function = partial(func, filelock = filelock, **kwargs)

        # this needs to be invoked by iterating over it
        results = [result for result in p.imap(map_function, chunks)]
    
    return results
    
            
def singleprocess_map(chunks: Iterable, func: Callable, **kwargs) -> None:
    """
    processes chunks with func using map. single threaded equivalent to multiprocess_map
    
    :param chunks:        Iterable containing chunks of a whole that need to be processed concurrently
    :param func:          function to process the chunks with
    :param **kwargs:      any keyword arguments to pass to func
    
    :return:              None
    """
    map_function = partial(func, **kwargs)
    
    results_iterable = map(map_function, chunks)
    # this needs to be invoked by iterating over it
    return list(results_iterable)


def process_data_in_chunks(iterable: Iterable, func: Callable, chunksize: int = 5000, n_processes: int = 1, **kwargs) -> Any:
    """
    uses func to process the given iterable in chunks of size chunksize possibly concurrently
    
    :param iterable:      iterable containing the data to be processed with func
    :param func:          function that processes the contents of iterable
    :param chunksize:     size of the indiviually processed chunks of the input iterable
    :param n_processes:   number of processes to use for mapping if n_processes > 1 this will be done concurrently using multiprocessing.imap
    :param **kwargs:      any keyword arguments that need to be passed to func
    
    :return:              anything that is returned by func
    """
    chunks = it.batched(
        iterable,
        n = chunksize
    )

    if n_processes > 1:
        results = multiprocess_map(
            chunks,
            func,
            n_processes,
            **kwargs
        )

    else:
        results = singleprocess_map(
            chunks,
            func,
            **kwargs
        )

    return results


import regex

import pandas as pd

from unidecode import unidecode
from typing import Iterable


def remove_whitespace(string: str) -> str:
    """
    removes whitespace from string

    :param string:      string object from which to remove whitespaces

    :return:            string without whitespace
    """
    no_whitespace_string = []
    for char in string:
        if char != ' ':
            no_whitespace_string.append(char)
    
    return ''.join(no_whitespace_string)


def all_equal(x: pd.Series) -> bool:
    """
    checks if all items in a pandas.Series are the same

    :param x:       pandas.Series object to check

    :return:        True if all items are the same else False
    """
    ref = unidecode(remove_whitespace(x.iloc[0]))
    return all(unidecode(remove_whitespace(item)) == ref for _, item in x.items())


def remove_empty_string_tokens(tokens: list[str]) -> list[str]:
    return [token for token in tokens if token]


def normalize_term(term: str, first_token_replacement: str) -> str:
    """
    this function fixes spelling mistakes and relies on the 
    assumption that term contains the word to be replaces at 
    the first position of the word sequence. Additionally, also
    replaces '-', '/' and '_' with ' '

    :param term:                        string containing the term
    :param first_token_replacement:     string used as a replacement for the first word in term

    :return:                            string containing the normalize term
    """
    replace_strings = ['-', '/', '_']
    for replace_string in replace_strings:
        term = term.replace(replace_string, ' ')
        
    tokens = remove_empty_string_tokens(term.split(' '))
    tokens[0] = first_token_replacement
    return ' '.join(tokens)


def extract_treatment(raw_metadata_items: list[str], matchers: dict[regex.Pattern]) -> str:
    """
    extracts all normalized terms and their values that match any of the regex patterns in matchers

    :param raw_metadata_items:  list of strings of the form "term: value"
    :param matchers:            dictionary containing first_token_replacement as key and regex.Patterns for matching as values

    :return:                    string of matched "term: value" pairs concatenated with '; ' or empty string if nothing matches
    """
    extracted_items = []
    for raw_metadata_item in raw_metadata_items:
        for term_name, term_matcher in matchers.items():
            m = term_matcher.match(raw_metadata_item)
            if not m:
                continue

            term, value = raw_metadata_item.split(': ')
            term = normalize_term(term, term_name)
            extracted_items.append(f'{term}: {value}')

            # guards against possible double matches 
            # which technically should not occur
            break
    
    return '; '.join(extracted_items)


def map_treatment(raw_sample_metadata: Iterable[str]) -> list[str]:
    # the regular expressions in this dictionary were built by non-exhaustive
    # trial and error and thus might not match everything that is contained in
    # the raw metadata but should match most of it. This trade-off is made
    # with the hope that the shared embedding of the model that will be trained on this
    # will anyway bring these samples together
    term_matchers = {
        'treatment': regex.compile('^t(reatm){i<=1,d<=1,s<=1,2i+2d+1s<=4}'),
        'transfection': regex.compile('^t(ransfect){i<=1,d<=1,s<=1,2i+2d+1s<=4}'),
        'transduction': regex.compile('^t(ransduc){i<=1,d<=1,s<=1,2i+2d+1s<=4}')
    }

    mapped_treatments = []
    for raw_metadata in raw_sample_metadata:
        raw_metadata_items = raw_metadata.split('; ')
        mapped_treatments.append(
            extract_treatment(raw_metadata_items, term_matchers)
        )
            
    return mapped_treatments



import requests
import logging 

import pandas as pd

from . import dbutils
from . import concurrency
from . import parsers


logging.basicConfig(
    format = '%(threadName)s: %(asctime)s-%(levelname)s-%(message)s',
    datefmt = '%Y-%m-%d %H:%M:%S',
    level = logging.INFO
)

# need to fetch from ftp because e utilities do not provide functionality
# can be done directly from archs4 using gsm and gse accessions
def fetch_soft_metadata(accession: str) -> list[str]:
    """
    fetches metadata from the GEO FTP server. Returns a (possibly) empty list

    :param accession:   string containing the GEO accession to fetch metadata for

    :return:            list of SOFT formatted strings if retrieval was successful else empty list
    """
    URL_BY_ACC = {
        "GSE": "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ=gse&acc={ACCESSION}&form=text&view=full",
        "GSM": "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ=gsm&acc={ACCESSION}&form=text&view=full",
    }
    url_base = URL_BY_ACC[accession[:3]]
    response = requests.get(
        url_base.format(ACCESSION = accession)
    )
    
    if not response.ok:
        logging.info(f'could not retrieve {accession}')
        return []

    response.encoding = "UTF-8"
    result_text = response.text
    result_list = result_text.replace("\r", "").split("\n")
    result_list = [elem for elem in result_list if len(elem) > 0]
        
    return result_list


def fetch_and_parse_gse_softs(
    gse_accessions: list[str], 
    already_fetched_softs: dict[str, list[str]]
) -> dict[str, str]:
    """
    fetches GEO Series metadata from GEO FTP. only returns the metadata of the
    oldest Series entry to make sure we retreive the original description.

    :param gse_accessions:          list of GEO series accessions as strings
    :param already_fetched_softs:   dictionary of already fetched GSE metadata to avoid unnecessary API calls

    :return:                        dictionary containing the metadata of interest of the oldest GEO series retrieved
    """
    gse_keys = {
        'Series_summary': 'series_summary',
        'Series_overall_design': 'series_design',
        'Series_submission_date': 'series_submission_date'
    }
    original_series_metadata = {}
    oldest_submission_date = None
    for gse in gse_accessions:
        if gse in already_fetched_softs:
            gse_soft = already_fetched_softs[gse]
            
        else:
            gse_soft = dbutils.retry(
                fetch_soft_metadata,
                gse
            )
            already_fetched_softs[gse] = gse_soft
        
        if parsers.is_superseries(gse_soft):
            continue
        
        gse_metadata = parsers.parse_soft_metadata(gse_soft, gse_keys)
        this_submission_date = parsers.to_date(
            gse_metadata['series_submission_date']
        )
        gse_metadata['series_submission_date'] = this_submission_date
        
        if not original_series_metadata:
            original_series_metadata = gse_metadata
            oldest_submission_date = this_submission_date
            continue
        
        if this_submission_date < oldest_submission_date:
            original_series_metadata = gse_metadata
            
    for key, value in original_series_metadata.items():
        if isinstance(value, list):
            original_series_metadata[key] = ''.join(value)
    
    return original_series_metadata


def fetch_geo_metadata(geo_accessions: list[tuple[str, str]]) -> pd.DataFrame:
    """
    retrieves the metadata for a list of GEO accessions and 
    returns it as a pandas.DataFrame

    :param geo_accessions:  list of GSM, GSE accession tuples. 
                            GSE can be a string of multiple GSE accessions separated by a ','

    :return:                pandas.DataFrame of retrieved metadata
    """
    gsm_keys = {
        'Sample_treatment_protocol_ch1': 'treatment_protocol',
        'Sample_growth_protocol_ch1': 'growth_protocol'
    }

    metadata = {}
    already_fetched_gses = {}
    for gsm, gse in geo_accessions:
        gsm_soft = dbutils.retry(
            fetch_soft_metadata,
            gsm
        )
        gsm_metadata = parsers.parse_soft_metadata(gsm_soft, gsm_keys)
        gsm_metadata.update(
            fetch_and_parse_gse_softs(
                gse.split(','),
                already_fetched_gses
            )
        )
        metadata[gsm] = gsm_metadata

    metadata = pd.DataFrame.from_dict(
        metadata,
        orient = 'index'
    )

    return metadata


def geo_metadata(
    geo_accessions: list[tuple[str, str]], 
    n_processes: int = 1,
    chunksize = 10000
) -> pd.DataFrame:
    """
    retrieves the metadata for a list of GEO accessions and 
    returns it as a pandas.DataFrame. More or less a concurrent version
    of `fetch_geo_metadata`

    :param geo_accessions:  list of GSM, GSE accession tuples. 
                            GSE can be a string of multiple GSE accessions separated by a ','
    :param n_processes:     how many concurrent processes to use to retrieve the data
    :param chunksize:       size of the chunks that are used for each process.

    :return:                pandas.DataFrame of retrieved metadata
    """
    
    metadata_frames = concurrency.process_data_in_chunks(
        geo_accessions,
        fetch_geo_metadata,
        n_processes = n_processes,
        chunksize = chunksize
    )

    return pd.concat(metadata_frames)



import certifi
import urllib3
import json

import pandas as pd
import itertools as it

from typing import Iterable, Union
from os import PathLike
from unidecode import unidecode
from . import obo
from io import BytesIO


def study_id_to_metasra(study_ids: Iterable[str]) -> pd.DataFrame:
    """
    used the MetaSRA API to retrieve normalized metdata for all samples in 
    a study given by study_ids from their database
    
    :param study_ids:   iterable containing SRA Biostudy accessions

    :return:            pandas.DataFrame containing MetaSRA normalised metadata
    """
    metasra_data = []
    for i, study_id in enumerate(study_ids):
        # this was necessary before metasra switched to https
        # http = urllib3.PoolManager(
        #     ca_certs = certifi.where(),
        #     cert_reqs = 'CERT_NONE'                      
        # )

        # retrieving number of studies that match search criteria
        r = urllib3.request(
            'GET', 
            'https://metasra.biostat.wisc.edu/api/v01/samples.csv',
            fields = {
                'study': study_id,
                'species': 'human', 
                'assay': 'RNA-seq', 
                'limit': 10000,
                'skip': 0
            }
        )

        study_data = pd.read_csv(BytesIO(r.data))
        
        if not study_data.empty:
            metasra_data.append(study_data)
    
    return pd.concat(metasra_data)


def convert_attributes_to_dict(attributes: str) -> dict[str, str]:
    """
    converts the attributes string into an attributes dictionary

    :param attributes:  string of key-value pairs of the format key1:value1;key2:value2; ...

    :return:            dictionary of attributes with key as key and value as value
    """
    attribute_dict = dict()
    for kv_pair in attributes.split('; '):
        k, v = kv_pair.split(': ')
        attribute_dict[k] = v
    
    return attribute_dict


def convert_to_dict_and_write_json(metadata_rows: Iterable[pd.Series], output_json: Union[PathLike, str]) -> None:
    """
    takes an iterable of pandas.DataFrame rows as pd.Series with keys 'accession' and 'attribute' converts them into
    a dictionary with key accession and attribute dictionary as values and writes this as JSON to output_json

    :param metadata_rows:   iterable of pd.Series with keys 'accession' and 'attribute'
    :param output_json:     name of the outputfile to write the JSON formatted data to
    
    :return:                None
    """
    json_dict = {
        row.accession: convert_attributes_to_dict(row.attribute) for _, row in metadata_rows
    }
    with open(output_json, 'w') as jsonfile:
        json.dump(json_dict, jsonfile, indent=4, separators=(',', ': '))


def raw_metadata_to_json(metadata: pd.DataFrame, output_json: Union[PathLike, str], chunksize = -1) -> None:
    """
    takes a pandas.DataFrame with columns 'accession' and 'attribute' and converts it to a
    MetaSRA input JSON style dictionary (accession as key, dictionary of attributes as value)
    which is then written to a JSON file. If chunksize > 0 the DataFrame is split into chunks
    and multiple outputfiles will be created.

    :param metadata:        pandas.DataFrame with columns 'accession' and 'attribute'
    :param output_json:     name of the output JSON file, will be appended with numbers 1 - n_chunks if chunksize > 0
    :param chunksize:       number of individual samples per chunk, n_chunks = n_samples / chunksize

    :return:                None
    """
    output_file_prefix = '.'.join(output_json.split('.')[:-1])
    output_file_suffix = output_json.split('.')[-1]
    if chunksize > 0:
        chunks = it.batched(
            metadata.iterrows(),
            chunksize
        )
        for i, chunk in enumerate(chunks):
            convert_to_dict_and_write_json(
                chunk, 
                f'{output_file_prefix}_{i+1}.{output_file_suffix}'
            )

    else:
        convert_to_dict_and_write_json(
            metadata.iterrows(),
            f'{output_file_prefix}.{output_file_suffix}'
        )


def parse_real_value_property(real_value_property: dict[str, str], ontologies: list[dict[str, str]]) -> tuple[str, str]:
    """
    parses the given real value property dictionary and returns
    two string representation one for the contained IDs and one
    for the corresponding terms

    :param real_value_property:     dictionary containing keys 'property_id', 'unit_id' and 'value'
    :param ontologies:              list of dictionaries containing id to term mapping for all ontologies used

    :return:                        string representation of real value property with ontology ids and ontology terms
    """
    property_id = real_value_property['property_id']
    property_unit_id = real_value_property['unit_id']
    property_value = real_value_property['value']

    property_term = obo.map_id_to_term(property_id, ontologies)
    property_unit_term = (
        'missing' if property_unit_id == 'missing' 
        else obo.map_id_to_term(property_unit_id, ontologies)
    )
    return f'{property_id}: {property_value}[{property_unit_id}]', f'{property_term}: {property_value}[{property_unit_term}]'


def metasra_output_json_to_dataframe(output_json: Union[PathLike, str], ontologies: list[dict[str, str]]) -> pd.DataFrame:
    """
    parses the output of MetaSRA into a pandas.DataFrame

    :param output_json:     string denoting the path to the output JSON file to parse
    :param ontologies:      list of it_to_term mappings used for mapping real value property ids to terms

    :return:                pandas.DataFrame containing the output in the following columns 
                            'accession', 
                            'sample_type', 
                            'sample_type_confidence', 
                            'mapped_ontology_ids', 
                            'mapped_ontology_terms',
                            'real_value_property_ids',
                            'real_value_property_terms'
    """
    with open(output_json, 'r') as f:
        mappings = json.load(f)

    metasra_data = []
    for accession, mapping in mappings.items():
        sample_type = mapping['sample type']
        sample_type_confidence = mapping['sample-type confidence']
        mapped_ontology_terms, mapped_ontology_ids = [], []
        for term in mapping['mapped ontology terms']:
            term_id, term_name = term.split('|')
            mapped_ontology_terms.append(term_name)
            mapped_ontology_ids.append(term_id)

        real_value_ids, real_value_terms = [], []
        for real_value_property in mapping['real-value properties']:
            real_value_id, real_value_term = parse_real_value_property(
                real_value_property,
                ontologies
            )
            real_value_ids.append(real_value_id)
            real_value_terms.append(real_value_term)

        metasra_data.append(
            [
                accession, 
                sample_type, 
                sample_type_confidence, 
                ', '.join(mapped_ontology_ids), 
                ', '.join(mapped_ontology_terms),
                ', '.join(real_value_ids),
                ', '.join(real_value_terms)
            ]
        )

    metasra_df = pd.DataFrame(
        metasra_data, 
        columns = [
            'accession', 
            'sample_type', 
            'sample_type_confidence', 
            'mapped_ontology_ids', 
            'mapped_ontology_terms',
            'real_value_property_ids',
            'real_value_property_terms'
        ]
    )
    return metasra_df
        

def make_accession_and_attributes_table(archs4_annotated: pd.DataFrame, accession_column: str = 'sample') -> pd.DataFrame:
    """
    takes an annotated ARCHS4 metdata table as returned by metadata.merge_to_annotated_metadata
    and extracts a given accession column and the raw BioSample attributes. removes samples
    that do not have any attributes

    :param archs4_annotated:    annotated ARCHS4 metadata table (see metadata.merge_to_annotated_metadata)
    :param accession_column:    string denoting the column with the desired NCBI accession

    :return:                    pandas.DataFrame with columns 'accession' and 'attribute'
    """
    accession_and_attributes = archs4_annotated.loc[
        :, 
        [accession_column, 'raw_biosample_metadata']
    ].rename(
        columns = {
            accession_column: 'accession',
            'raw_biosample_metadata': 'attribute'
        }
    ).dropna()

    # map unicode characters to ascii
    accession_and_attributes.loc[:, 'attribute'] = accession_and_attributes.attribute.apply(
        unidecode
    )

    return accession_and_attributes.drop_duplicates()



import logging

import pandas as pd

from . import parsers
from . import summary
from . import concurrency
from . import dbutils
from . import metasra
from . import metadatautils
from typing import Union, Iterable
from os import PathLike

logging.basicConfig(
    format = '%(threadName)s: %(asctime)s-%(levelname)s-%(message)s',
    datefmt = '%Y-%m-%d %H:%M:%S',
    level = logging.INFO
)

def map_accessions_to_srauids(
    table: pd.DataFrame, 
    outfilename: Union[PathLike, str], 
    chunksize: int = 5000, 
    n_processes: int = 1
) -> None:
    """
    takes a pandas.DataFrame containing columns srx_accession, biosample_accession and geo_accession, maps those accessions
    to SRA UIDs and writes the resulting map to outfilename. If n_processes > 1 this will be done concurrently
    
    :param table:         pandas.DataFrame containing columns srx_accession, biosample_accession and geo_accession
    :param db:            string denoting the NCBI database to retrieve the UIDs from
    :param outfilename:   path to the outputfile the resulting map should be written to
    :param chunksize:     size of the indiviually processed chunks of the input table
    :param n_processes:   number of processes to use for mapping if n_processes > 1 this will be done concurrently using multiprocessing.imap
    
    :return:              None
    """
    mapping_table = parsers.get_not_yet_mapped(
        table,
        outfilename
    )
    
    concurrency.process_data_in_chunks(
        mapping_table.iterrows(),
        summary.map_accessions_to_uids,
        db = 'sra',
        outfilename = outfilename,
        n_processes = n_processes,
        chunksize = chunksize
    )


def determine_accession_type(accession: str) -> str:
    """
    determines the type of accession used for retrieving the UIDs
    currently identifies SRX SAMN or GSM accessions

    :param accession:   string containing the accession to be identified

    :return:            string giving the type of accession
    """
    types = {
        'experiment': 'SRX',
        'biosample': 'SAMN',
        'geo': 'GSM'
    }
    
    for accession_type, accession_prefix in types.items():
        if accession.startswith(accession_prefix):
            return accession_type
        

def merge_uids_and_accessions(uids: pd.DataFrame, accessions: pd.DataFrame) -> pd.DataFrame:
    """
    merges the uid frame with the retrieved accessions

    :param uids:            pandas.DataFrame containing 'uid' and 'accession' columns
    :param accessions:      pandas.DataFrame containing accessions correspondig to uids

    :return:                pandas.DataFrame merged on uid
    """
    uids = uids.copy()
    uids['accession_type'] = uids.accession.apply(
        determine_accession_type
    )

    merged_groups = []
    for accession_type, group in uids.groupby('accession_type'):
        group = group.rename(
            columns = {'accession': accession_type}
        )
        
        merged_group = group.merge(
            accessions,
            on = accession_type,
            how = 'inner'
        )

        merged_groups.append(merged_group)
        
    return pd.concat(merged_groups)


def srauids_to_accessions(srauids: pd.DataFrame, chunksize: int = 50000) -> pd.DataFrame:
    """
    retrieves all available accessions for the sra uids given in 'uid' column of srauids

    :param srauids:     pandas.DataFrame containing an 'accession' and a 'uid' column
    :param chunksize:   integer giving the size of the chunks posted and retrieved from Entrez

    :return:            pandas.DataFrame containing experiment, sample, biosample, study, bioproject and geo accessions
    """
    accessions = summary.summaries_from_uids(
        srauids.uid,
        'sra',
        parsers.accessions_from_esummary_response,
        parsers.accession_matchers,
        chunksize = chunksize
    )
    
    return merge_uids_and_accessions(srauids, accessions)


def biosample_uids_to_metadata(biosample_uids: Iterable[Union[str, int]], chunksize: int = 50000) -> pd.DataFrame:
    """
    retrieves all available metadata for the biosample uids

    :param biosample_uids:  iterable of biosample UIDs as string or int
    :param chunksize:       integer giving the size of the chunks posted and retrieved from Entrez

    :return:            pandas.DataFrame containing experiment, sample, biosample, study, bioproject and geo accessions
    """
    metadata = summary.summaries_from_uids(
        biosample_uids,
        'biosample',
        parsers.metadata_from_biosample_uids,
        parsers.biosample_metadata_parsers,
        chunksize = chunksize
    )
    
    return metadata


def metasra_from_study_id(study_ids: Iterable[str]) -> pd.DataFrame:
    """
    used the MetaSRA API to retrieve normalized metdata for all samples in 
    a study given by study_ids from their database
    
    :param study_ids:   iterable containing SRA Biostudy accessions

    :return:            pandas.DataFrame containing MetaSRA normalised metadata
    """
    metasra_data = dbutils.process_in_chunks(
        study_ids,
        metasra.study_id_to_metasra,
        chunksize = 1000
    )

    return pd.concat(metasra_data)


def merge_to_annotated_metadata_frame(
    biosample_metadata: pd.DataFrame, 
    srauids_to_biosampleuids: pd.DataFrame,
    ncbi_accessions: pd.DataFrame,
    archs4_metadata: pd.DataFrame
) -> pd.DataFrame:
    """
    combines the passed pandas.DataFrames to a fully annotated ARCHS4 metadata set

    :param biosample_metadata:          pandas.DataFrame containing BioSample metadata (see biosample_uids_to_metadata)
    :param srauids_to_biosampleuids:    pandas.DataFrame containing the linked SRA to BioSample UIDs (see link.link_sra_to_biosample)
    :param ncbi_accessions:             pandas.DataFrame containing the SRA accessions (see srauids_to_accessions)
    :param archs4_metadata:             pandas.DataFrame containing the extracted ARCHS4 metadata (see archs4.get_filtered_sample_metadata)

    :return:                            pandas.DataFrame containing annotated ARCHS4 metadata
    """
    biosample_uids_and_metadata = srauids_to_biosampleuids.merge(
        biosample_metadata,
        on = 'biosample',
        how = 'inner'
    )

    accessions_biosample_metadata_uids = biosample_uids_and_metadata \
        .rename(
            columns = {
                'sra': 'sra_uid',
                'biosample': 'biosample_uid',
                'title': 'biosample_title',
                'attribute': 'raw_biosample_metadata'
            }
        ).merge(
            ncbi_accessions.rename(
                columns = {
                    'uid': 'sra_uid'
                }
            ),
            on = 'sra_uid',
            how = 'inner'
        )

    archs4_annotated = accessions_biosample_metadata_uids.merge(
        archs4_metadata.drop(
            columns = [
                    'srx_accession', 
                    'biosample_accession'
                ]
        ).rename(
            columns = {
                'geo_accession': 'geo',
                'title': 'geo_title',
                'source_name_ch1': 'geo_source_name',
                'characteristics_ch1': 'geo_metadata'
            }
        ),
        on = 'geo',
        how = 'inner'
    )

    # this is done in order to remove scrambled annotations due to spurious links to BioSample
    spurious_duplication = archs4_annotated.sra_uid.duplicated(keep = False)
    archs4_annotated_nonduplicated = archs4_annotated.loc[~spurious_duplication, :]
    archs4_annotated_duplicated = archs4_annotated.loc[spurious_duplication, :]

    all_equal_idx = archs4_annotated_duplicated[['geo_title', 'biosample_title']].apply(
        metadatautils.all_equal,
        axis = 1
    )

    archs4_annotated_deduplicated = pd.concat(
        [
            archs4_annotated_nonduplicated,
            archs4_annotated_duplicated.loc[all_equal_idx, :]
        ]
    )
    return archs4_annotated_deduplicated.reset_index(drop = True)


def annotated_with_metasra_and_treatment(annotated_frame: pd.DataFrame, metasra_frame: pd.DataFrame) -> pd.DataFrame:
    """
    merges the annotated DataFrame returned by 'merge_to_annotated_metadata_frame' with the normalized metadata DataFrame
    produced by MetaSRA and as returned by 'metasra.metasra_output_json_to_dataframe'. Additionally, the function also 
    attempts to parse out any treatment information from the raw metadata which is usually not kept by metasra into a
    new column 'treatment'. This includes any raw metadata property that starts with 'treat', 'transfect' or 'transduc'.
    Note that resulting matches are normalized by replacing the first word in the match with either 'treatment', 'transfection'
    or 'transduction' respectively. This is a crude attempt to remove any spelling mistakes.

    :param annotated_frame:     pandas.DataFrame as returned by 'merge_to_annotated_metadata_frame'
    :param metasra_frame:       pandas.DataFrame as returned by 'metasra.metasra_output_json_to_dataframe'

    :return:                    merged pandas.DataFrame containing an additional new column 'treatment'
    """
    sample_accessions = annotated_frame.loc[:, ['sample', 'geo']]
    metasra_accession_annotated = metasra_frame.merge(
        sample_accessions.rename(columns = {'sample': 'accession'}),
        on = 'accession',
        how = 'inner'
    )
    metasra_accession_annotated.drop_duplicates(inplace = True)

    metasra_annotated = metasra_accession_annotated.merge(
        annotated_frame.rename(
            columns = {
                'geo_accession': 'geo',
                'title': 'geo_title',
                'source_name_ch1': 'geo_source_name',
                'characteristics_ch1': 'geo_metadata'
            }
        ),
        on = 'geo',
        how = 'inner'
    )

    metasra_annotated['treatment'] = metadatautils.map_treatment(
        metasra_annotated.raw_biosample_metadata
    )

    return metasra_annotated



import re
import datetime
import os

import pandas as pd

from xml.etree import ElementTree
from typing import Any, Union, Callable
from os import PathLike


MONTHS = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
MONTH_TO_INT = {
    month_str: i for i, month_str in enumerate(MONTHS, 1)
}


def node_parser(node: ElementTree.Element, parse_func: Callable, match_string: str) -> tuple[bool, dict[str, str]]:
    """
    parses an xml node if with parse_func if node.tag == match_sting
    and returns True and a dictionary with the nodes attributes
    else returns False and an empty dictionary

    :param node:            ElementTree.Element instance
    :param parse_func:      function taking the node and parses out its information
    :param match_string:    string representing the nodes tag that needs to match for it to be parsed

    :return:                True, dictionary with parsed infos if match_string == node.tag else False, empty dictinary   
    """
    match = False
    if not node.tag == match_string:
        return match, dict()
    
    match = True
    return match, parse_func(node)


def parse_title(node: ElementTree.Element) -> dict[str, str]:
    """
    returns the the text attribute of node if node.tag == 'title'

    :param node:    ElementTree.Element instance

    :return:        dictionary containing the title
    """
    return {'title': node.text}
        
    
def parse_attribute(node: ElementTree.Element) -> dict[str, str]:
    """
    returns the nodes attributes if node.tag == 'attribute'
    keys of these attributes are prepended with 'attribute|'
    for easier downstream processing

    :param node:    ElementTree.Element instance

    :return:        dictionary containing the nodes attributes
    """
    harmonized_name = node.attrib.get('harmonized_name')
    attribute_name = node.attrib.get('attribute_name')
    name = harmonized_name if harmonized_name else attribute_name
    return {f'attribute|{name}': node.text}


def parse_organism(node: ElementTree.Element) -> dict[str, str]:
    """
    returns the organism and taxonomy id if node.tag == 'organism'

    :param node:    ElementTree.Element instance

    :return:        dictionary containing the organism and taxonomy id
    """
    return {'organism': ';'.join(node.attrib[k] for k in ['taxonomy_name', 'taxonomy_id'])}


def parse_biosample_uid(node: ElementTree.Element) -> dict[str, str]:
    """
    returns the biosample accession if node.tag == 'biosample'

    :param node:    ElementTree.Element instance

    :return:        dictionary containing the biosample id
    """
    return {'biosample': node.attrib['id']}


biosample_metadata_parsers = {
    'Title': parse_title,
    'Attribute': parse_attribute,
    'Organism': parse_organism,
    'BioSample': parse_biosample_uid
}


def combine_attribute_keys(metadata: dict[str, Any]) -> dict[str, Any]:
    """
    extracts all keys starting with 'attribute|' from metadata
    and puts them back as a single 'attribute' key joining the
    individual key-value-pairs to string as follows "key: value; key: value; ..."

    :param metadata:    dictionary containing the metadata for a given sample parses from XML

    :return:            metadata with a single attribute key added back instead of all "attribute|" keys
    """
    attribute_kv_list = []
    revised_metadata = dict()
    for key, value in metadata.items():
        if not key.startswith('attribute'):
            revised_metadata[key] = value
            continue
            
        name = key.split('|')[-1]
        value = value.replace(';', '')
        value = value.replace(':', '')
        attribute_kv_list.append(f'{name}: {value}')
    
    revised_metadata['attribute'] = '; '.join(attribute_kv_list)
    return revised_metadata


def extract_metadata(xmltree: ElementTree, metadata_parsers: dict[str, Callable]) -> dict[str, Any]:
    """
    parses the XML formated metadata and returns a dictionary of it

    :param xmltree:             ElementTree instance initialized with the biosample XML response from eSummary
    :param metadata_parsers:    dictionary containing matching node tags as keys and node parser functions as values

    :return:                    dictionary containing the metadata parsed with metadata_parsers
    """
    metadata = dict()
    for node in xmltree.iter():
        for match_string, parse_function in metadata_parsers.items():
            match, parse_result = node_parser(
                node, 
                parse_function, 
                match_string
            )
            
            if match:
                metadata.update(parse_result)
                break
                
    return combine_attribute_keys(metadata)


def add_parsing_results_to_metadata(parsed_metadata: dict[str, Any], metadata: dict[str, list[Any]]) -> None:
    """
    adds parsed_metadata values to the list under the corresponding key in metadata
    this is done in place

    :param parsed_metadata:     the parsed metadata of a single eSummary response
    :param metadata:            a collection of parsed metadata from multiple eSummary responses

    :return:                    None
    """
    for key in metadata.keys():
        metadata[key].append(
            parsed_metadata.get(key)
        )
        

def metadata_from_biosample_uids(
    summary_items: list[dict[str, Any]],
    metadata: dict[str, list[Any]], 
    metadata_parsers: dict[str, Callable]
) -> None:
    """
    parses the sSummary responses from multiple items and adds it to the metadata dictionary in place

    :param summary_items:       esummary response as returned by Entrez.read
    :param metadata:            dictionary the parsed metadata should be stored in
    :param metadata_parsers:    dictionary containing node.tags as keys and corresponding parser functions as values

    :return:                    None
    """
    for summary in summary_items['DocumentSummarySet']['DocumentSummary']:
        xmltree = ElementTree.fromstring(summary['SampleData'])
        parsed_metadata = extract_metadata(
            xmltree,
            metadata_parsers
        )
        add_parsing_results_to_metadata(
            parsed_metadata,
            metadata
        )


def extract_accessions(
    summary: dict[str, Any], 
    accession_summary: dict[str, list], 
    accession_matchers: dict[str, re.Pattern]
) -> None:
    """
    parses the accessions specified in accession_matchers and adds them to the accession_summary dictionary in place

    :param summay:                  summary item returned by Bio.Entrez.esummary
    :param accession_summary:       dictionary containing the same keys ad accession_matchers and lists as values
    :param accession_matchers:      dictionary containing re.Pattern as values

    :return:                        None
    """
    for accession_type, accession_matcher in accession_matchers.items():
        match = accession_matcher.search(summary['ExpXml'])
        accession = match.groups()[0] if match else None
        accession_summary[accession_type].append(accession)


def accessions_from_esummary_response(
    summary_items: list[dict[str, Any]], 
    accessions: dict[str, list], 
    accession_matchers: dict[str, re.Pattern]
) -> None:
    """
    parses the summary items from eSummary

    :param summary_items:           list of summary items as returned by Bio.Entrez.esummary
    :param accession_summary:       dictionary containing the same keys ad accession_matchers and lists as values
    :param accession_matchers:      dictionary containing re.Pattern as values

    :return:                        None
    """
    for summary in summary_items:
        extract_accessions(
            summary,
            accessions,
            accession_matchers
        )


accession_matchers = {
    'experiment': re.compile('Experiment acc="([SRX0-9]+)'),
    'sample': re.compile('Sample acc="(SRS[0-9]+)'),
    'study': re.compile('Study acc="(SRP[0-9]+)'),
    'biosample': re.compile('Biosample>(SAMN[0-9]+)'),
    'bioproject': re.compile('Bioproject>(PRJNA[0-9]+)'),
    'geo': re.compile('name="(GSM[0-9]+)')
}


def read_retrieved_accessions(filename: Union[PathLike, str]) -> set[str]:
    """
    reads the file given with filename and return a list of unique accessions
    the file has to contain a table with an 'accession' column
    
    :param filename:   path to the accession table to read
    
    :return:           set of unique accessions
    """
    accession_map = pd.read_csv(
        filename,
        sep = '\t'
    )
    return set(accession_map.accession)


def get_not_yet_mapped(table: pd.DataFrame, outfilename: Union[PathLike, str]) -> pd.DataFrame:
    """
    If outfilename does not exist yet it is created and table is returned as is. If outfilename exists
    accessions contained in it are already mapped and thus will be removed from table before returning it
    
    :param table:         pandas.DataFrame containing the accessions we want to retrieve
    :param outfilename:   path to a file the mapping output should be or is written to
    
    :return:              pandas.DataFrame containing only the accessions that are not already found in outfilename
    """
    # this can definitely be refactored but I leave it as is for now
    if not os.path.exists(outfilename):
        retrieved_accessions = set()
        with open(outfilename, 'w') as outfile:
            outfile.write(
                'accession\tuid\tdatabase\n'
            )
            
        return table

    else:
        retrieved_accessions = read_retrieved_accessions(outfilename)
        retrieved = table.apply(
            lambda x: any(
                x[acc] in retrieved_accessions 
                for acc in ['geo_accession', 'srx_accession', 'biosample_accession']
            ),
            axis = 1
        )
        return table.loc[~retrieved, :]
    

def parse_soft_line(line: str) -> tuple[str, str]:
    """
    parse line of SOFT formatted metadata

    :param line:    SOFT formatted string

    :return:        
    """
    key, value = line.split(' = ', maxsplit = 1)
    return key[1:], value


def to_list(item: Union[str, list]) -> list[str]:
    if isinstance(item, list):
        return item
    
    return [item]


def parse_soft_metadata(
    soft_metadata: list[str], 
    retain_keys: dict[str, str] = {}
) -> dict[str, str]: 
    """
    parses a list of SOFT formatted strings and returns a dictionary
    containing the parsed information. If retain_keys is given, only retains
    those the keys contained as keys and replaces them with value in the final
    result (e.g. retain_keys = {'key': 'other_key'} -> result = {'other_key': <metadata>})

    :param soft_metadata:   list of SOFT formatted strings
    :param retain_keys:     dictionary containing the keys to retain as keys and
                            the renamed key as values. result will contain renamed key
    
    :return:                dictionary containing the parsed metadatas
    """
    retain_all = True if not retain_keys else False
    metadata_dict = {}
    for line in soft_metadata:
        if not line.startswith('!'):
            continue
            
        key, value = parse_soft_line(line)
        if key not in retain_keys and not retain_all:
            continue
        
        if not retain_all:
            key = retain_keys[key]
            
        if key in metadata_dict:
            tmp = to_list(metadata_dict[key])
            tmp.append(value)
            metadata_dict[key] = tmp
        
        else:
            metadata_dict[key] = value
    
    return metadata_dict


def is_superseries(gse_soft: list[str]) -> bool:
    """
    determines if a given series is a superseries

    :param gse_soft:    list of SOFT formatted strings

    :return:            True if any line contains "SuperSeries" else False
    """
    regex = re.compile('SuperSeries')
    for line in gse_soft:
        if regex.search(line):
            return True
    
    return False


def to_date(date: str) -> datetime.date:
    """
    convert date string to date object for easier comparison

    :param date:    string containing the date

    :return:        date as datetime.date object
    """
    month, day, year = date.split()
    return datetime.date(int(year), MONTH_TO_INT[month], int(day))


import logging

from Bio import Entrez
from typing import Any, Union, Iterable
from . import dbutils
from functools import partial

import pandas as pd
import itertools as it


def has_links(link_item: dict[str, Any]) -> bool:
    """
    checks retrieved items for presence of any links

    :param link_item:   dictionary as returned by Bio.Entrez.read

    :return:            True if key 'LinkSetDb' is not an empty list False otherwise
    """
    return True if link_item['LinkSetDb'] else False


def get_links(link_item: dict[str, Any]) -> list[list[str, str]]:
    """
    parses the given link item and retruns all possible links
    as all combinations of the source id and linked ids

    :param link_item:   dictionary as returned by Bio.Entrez.read

    :return:            list of lists containing the respective links
    """
    # some accessions are not released so we can't link them
    if has_links(link_item):
        dbto_ids = link_item['LinkSetDb'][0]['Link']
        
    else:
        dbto_ids = [{'Id': None}]
        
    dbfrom_ids = link_item['IdList']
    
    links = [
        [dbfrom_id, dbto_id['Id']] for dbfrom_id, dbto_id in it.product(dbfrom_ids, dbto_ids)
    ]
    
    return links
    

def link_uids(uid_list: list[Union[str, int]], dbfrom: str, dbto: str) -> pd.DataFrame:
    """
    takes an iterable of UIDs originating from dbfrom and links them to dbto via eLink

    :param uid_list:    list of integers or strings denoting the UIDs from dbfrom
    :param dbfrom:      string denoting the NCBI database the UIDs stem from
    :param dbto:        string denoting the NCBI database to link the UIDs to

    :return:            pandas.DataFrame with columns dbfrom, dbto containing the linked UIDs
    """
    response_handle = Entrez.elink(
        dbfrom = dbfrom,
        db = dbto,
        id = uid_list
    )

    response = Entrez.read(response_handle)

    linked_ids = []
    for link_item in response:
        links = get_links(link_item)
        linked_ids.extend(links)
    

    linked_ids = pd.DataFrame(
        linked_ids,
        columns = [dbfrom, dbto]
    )
    
    return linked_ids


def link_sra_to_biosample(sra_uids: Iterable[Union[str, int]]) -> pd.DataFrame:
    """
    links SRA UIDs to BioSample UIDs

    :param sra_uids:    iterable containg string or integer versions of SRA UIDs to link

    :return:            pandas.DataFrame containing the 'sra' and 'biosample' column with the respective UIDs
    """
    retry_link = partial(
        link_uids,
        dbfrom = 'sra',
        dbto = 'biosample'
    )
    links = dbutils.process_in_chunks(
        sra_uids,
        retry_link,
        chunksize = 1000
    )
    return pd.concat(links)
    


import time
import logging

from Bio import Entrez
from functools import partial
from typing import Callable, Any, Union, Iterable
from os import PathLike

import multiprocessing as mp
import pandas as pd
import itertools as it

from . import concurrency
from . import dbutils


def uid_from_esearch(accession: str, db: str) -> list[list[str]]:
    """
    uses Bio.Entrez.esearch to request the database internal UID of the given accession
    see  https://www.ncbi.nlm.nih.gov/books/NBK25497/table/chapter2.T._entrez_unique_identifiers_ui/?report=objectonly
    for available databases
    
    :param accession:  accession to retrieve UID for
    :param db:         name of the Entrez database to retrieve the UID from
    
    :return:           list of lists of accession, uid and database name the uid was retrieved from
    """
    response_handle = Entrez.esearch(db = db, term = accession)
    uid_list = Entrez.read(response_handle)['IdList']
    return [[accession, uid, db] for uid in uid_list]


def map_accessions_to_uids(
    accessions: Iterable[pd.Series], 
    db: str, 
    outfilename: Union[PathLike, str], 
    filelock: Union[mp.Manager().Lock, None] = None
) -> None:
    """
    takes a list of pandas.Series object generated by pandas.DataFrame.iterrows and a path to an outputfile
    and uses uid_from_esearch to map the contained accessions to the SRA internal UID. If the function is used
    concurrently a Lock object must be passed via the filelock argument
    
    :param accessions:     list of pandas.Series objects containing the keys srx_accession, biosample_accession and geo_accession
    :param db:             string denoting the NCBI database to retrieve the UIDs from
    :param outfilename:    path to the file to write the retrieved UID mappings to
    :param filelock:       A Lock object used to savely write to the outfile in case of concurrent usage
    
    :return:               None
    """
    logging.info('starting mapping process')
    write_to_file = (
        concurrency.write_to_file_multiprocess if filelock 
        else concurrency.write_to_file_singleprocess
    )
    
    uid_list = []
    start = accessions[0][0]
    for i, row in accessions:
        srx = row.srx_accession
        samn = row.biosample_accession
        gsm = row.geo_accession
        
        if not any([srx, samn]):
            accession = gsm

        else:
            accession = srx if srx else samn
            
        mapped_uids = dbutils.retry(
            uid_from_esearch,
            accession, 
            db
        )
        
        uid_list.extend(mapped_uids)
        
        if not (i + 1) % 1000:
            write_to_file(
                outfilename,
                uid_list,
                filelock
            )
            uid_list = []
            logging.info(f'Written uids for accessions {start} to {i}')
    
    
    logging.info(f'finishing up. writing {len(uid_list)} remaining uid mappings')

    # write remaining uids if there are any
    if uid_list:
        write_to_file(
            outfilename,
            uid_list,
            filelock
        )
    

def data_from_esummary(
    web_env_info: dict[str, str], 
    db: str, 
    parse_function: Callable, 
    data_parsers: dict[str, Any], 
    n_uids: int, 
    chunksize: int = 5000
) -> pd.DataFrame:
    """
    retrieves all the data from eSummary associated with the UIDs posted to the Entrez history server

    :param web_env_info:        dictionary containing the keys 'QueryKey' and 'WebEnv' as returned by Entrez.epost
    :param db:                  string denoting the database the supplied UIDs are from
    :param parse_function:      function used to parse the esummary response
    :param data_parsers:        dictionary containing keys and whatever you parser needs to parse the response 
                                keys will be the columns of the returned dataframe
    :param n_uids:              number of UIDs posted to the history server
    :param chunksize:           size of the chunks of summaries to retrieve

    :return:                    pandas.DataFrame with columns equal to data_parsers keys and the parsed data as values
    """
    esummary = partial(
        Entrez.esummary,
        db = db,
        WebEnv = web_env_info['WebEnv'],
        query_key = web_env_info['QueryKey']
    )

    def esummaries_from_history(retstart, retmax):
        response_handle = dbutils.retry(
            esummary,
            retstart = retstart, 
            retmax = retmax
        )
        return Entrez.read(response_handle)
    
    data = {
        key.lower(): [] for key in data_parsers.keys()
    }
    
    for retmin, retmax in dbutils.get_chunklimits(n_uids, chunksize):
        time.sleep(5) # avoid too many requests error
        response = dbutils.retry(
            esummaries_from_history,
            retstart = retmin,
            retmax = retmax
        )
        parse_function(
            response,
            data,
            data_parsers
        )
        
    data = pd.DataFrame().from_dict(
        data,
        orient = 'columns'
    )
    
    # somehow we get duplicates, suspect the post call
    # because I did it twice so we might end up with 
    # duplicated uids in the webenv but not sure
    # anyway dropping duplicates here
    return data.drop_duplicates().reset_index(drop = True)


def uids_to_summaries(uid_list, db, parse_function, data_parsers):
    """
    retrieves summaries for uids from eSummary and returns a pandas.DataFrame with columns = keys of data_parsers
    Values in these columns are determined by parse_function

    :param uids:                iterable of uids to retrieve eSummaries for
    :param db:                  database to retrieve the summaries from
    :param parse_function:      function used to parse the summaries
    :param data_parsers:        parser dictionary forwarded to parse function
    :param chunksize:           size of the chunks of uids posted to the Entrez history server

    :return:                    pandas.DataFrame containing the summary parsing results
    """
    web_env_info = dbutils.create_webenv(uid_list, db)

    logging.debug(web_env_info)
    
    accessions = data_from_esummary(
        web_env_info,
        db,
        parse_function,
        data_parsers,
        len(uid_list)
    )

    return accessions


def summaries_from_uids(
    uids: Iterable[Union[int, str]], 
    db: str,
    parse_function: Callable, 
    data_parsers: dict[str, Any],
    chunksize: int = 50000
) -> pd.DataFrame:
    """
    retrieves summaries for uids from eSummary and returns a pandas.DataFrame with columns = keys of data_parsers
    Values in these columns are determined by parse_function

    :param uids:                iterable of uids to retrieve eSummaries for
    :param db:                  database to retrieve the summaries from
    :param parse_function:      function used to parse the summaries
    :param data_parsers:        parser dictionary forwarded to parse function
    :param chunksize:           size of the chunks of uids posted to the Entrez history server

    :return:                    pandas.DataFrame containing the summary parsing results
    """
    result_frames = dbutils.process_in_chunks(
        uids,
        uids_to_summaries,
        chunksize = chunksize,
        db = db,
        parse_function = parse_function,
        data_parsers = data_parsers,
    )

    return pd.concat(result_frames)
